;;; Disarchive
;;; Copyright © 2021, 2024 Timothy Sample <samplet@ngyro.com>
;;;
;;; This file is part of Disarchive.
;;;
;;; Disarchive is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Disarchive is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Disarchive.  If not, see <http://www.gnu.org/licenses/>.

(use-modules (disarchive kinds bzip2)
             (disarchive utils)
             (quickcheck)
             (quickcheck arbitrary)
             (quickcheck property)
             (rnrs bytevectors)
             (srfi srfi-64))

(define-syntax-rule (true-if-exception expr)
  (not (false-if-exception (or expr #t))))

(define tests-per-property 1000)

(configure-quickcheck
 (stop? (lambda (success-count _)
          (>= success-count tests-per-property))))

(test-begin "kinds--bzip2")


;; bzip2 stream headers

(define bzip2-magic-header #vu8(#x42 #x5a #x68))

(define $bzip2-stream-header
  ($record make-bzip2-stream-header
    (bzip2-stream-header-level $byte)))

(define (bytevector-set-bzip2-magic-header! bv)
  (bytevector-copy! bzip2-magic-header 0 bv 0
                    (min (bytevector-length bzip2-magic-header)
                         (bytevector-length bv))))

(test-assert "Refuses to decode a small bzip2 stream header"
  (true-if-exception
   (let ((bv (make-bytevector 3)))
     (bytevector-set-bzip2-magic-header! bv)
     (decode-bzip2-stream-header bv))))

(test-assert "Refuses to decode a large bzip2 stream header"
  (true-if-exception
   (let ((bv (make-bytevector 5)))
     (bytevector-set-bzip2-magic-header! bv)
     (decode-bzip2-stream-header bv))))

(test-assert "Refuses to decode bzip2 stream header without magic"
  (true-if-exception
   (let ((bv (make-bytevector 4)))
     (decode-bzip2-stream-header bv))))

(configure-quickcheck
 (size (const 4)))

(test-assert "[prop] Reading bzip2 stream headers is reversible"
  (quickcheck
   (property ((bv $bytevector))
     (bytevector-set-bzip2-magic-header! bv)
     (equal? bv (encode-bzip2-stream-header
                 (decode-bzip2-stream-header bv))))))

(test-assert "[prop] Writing bzip2 stream headers is reversible"
  (quickcheck
   (property ((header $bzip2-stream-header))
     (let ((bv (encode-bzip2-stream-header header)))
       (equal? header (decode-bzip2-stream-header bv))))))

(test-end "kinds--bzip2")
