;;; Disarchive
;;; Copyright © 2020, 2021 Timothy Sample <samplet@ngyro.com>
;;;
;;; This file is part of Disarchive.
;;;
;;; Disarchive is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Disarchive is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Disarchive.  If not, see <http://www.gnu.org/licenses/>.

(use-modules (disarchive kinds zero-string)
             (disarchive serialization)
             (quickcheck)
             (quickcheck arbitrary)
             (quickcheck property)
             (rnrs bytevectors)
             (srfi srfi-64)
             (tests kinds))

(configure-quickcheck
 ;; Perform 1000 tests per property...
 (stop? (lambda (success-count _)
          (>= success-count 1000)))
 ;; ...over input sizes from 0 to 10.
 (size (lambda (test-number)
         (if (zero? test-number)
             0
             (1+ (quotient test-number 100))))))

(test-begin "kinds--zero-string")

(test-equal "Ignores blank trailer"
  '("Hi" . "")
  (let ((zstr (decode-zero-string #vu8(72 105))))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-equal "Ignores trivial trailer"
  '("Hi" . "")
  (let ((zstr (decode-zero-string #vu8(72 105 0 0 0))))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-equal "Preserves non-trivial trailer"
  '("Hi" . "X")
  (let ((zstr (decode-zero-string #vu8(72 105 0 88))))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-equal "Handles blank string"
  '("" . "")
  (let ((zstr (decode-zero-string #vu8(0 0 0 0))))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-equal "Handles empty bytevector"
  '("" . "")
  (let ((zstr (decode-zero-string #vu8())))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-equal "Handles bytevector with one null byte"
  '("" . "")
  (let ((zstr (decode-zero-string #vu8(0))))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-equal "Handles bytevector with one non-null byte"
  '("X" . "")
  (let ((zstr (decode-zero-string #vu8(88))))
    (cons (zero-string-value zstr)
          (zero-string-trailer zstr))))

(test-assert "[prop] Reading is reversible"
  (quickcheck
   (property ((bv $bytevector))
     (let ((bv* (make-bytevector (bytevector-length bv))))
       (encode-zero-string (decode-zero-string bv) bv*)
       (equal? bv bv*)))))

(test-assert "[prop] Reading produces a valid result"
  (quickcheck
   (property ((bv $bytevector))
     (valid-zero-string? (decode-zero-string bv)))))

(test-assert "[prop] Writing is reversible"
  (quickcheck
   (property ((zstr $zero-string))
     (test-when (valid-zero-string? zstr)
       (equal? zstr (decode-zero-string (encode-zero-string zstr)))))))

(test-assert "[prop] Serializing is reversible"
  (quickcheck
   (property ((zstr $zero-string))
     (test-when (valid-zero-string? zstr)
       (equal? zstr (serdeser -zero-string- zstr))))))

(test-end "kinds--zero-string")
